#
#   5B - Controls
#

from pygame import Rect
from albow.widget import Widget
from albow.controls import Label, Button, TextField, Row, Column
from albow.dialogs import Dialog
from albow.resource import get_default_font

class TextAttrField(TextField):

	def __init__(self, width, client, attr = 'value', upper = False, **kwds):
		TextField.__init__(self, width = width, **kwds)
		self.client = client
		self.attr = attr
		self.upper = upper
		self.editing = False
	
	def get_text(self):
		if self.editing:
			return TextField.get_text(self)
		else:
			x = getattr(self.client, self.attr)
			return self.value_to_text(x)

	def set_text(self, x):
		if self.editing:
			TextField.set_text(self, x)
		else:
			setattr(self.client, self.attr, self.text_to_value(x))
	
	def value_to_text(self, x):
		return x
	
	def text_to_value(self, x):
		return x

	def insert_char(self, c):
		if self.upper:
			c = c.upper()
		if not self.editing:
			text = self.get_text()
			self.editing = True
			self.set_text(text)
		return TextField.insert_char(self, c)
	
	def enter_action(self):
		if self.editing:
			text = self.get_text()
			self.editing = False
			self.set_text(text)
			self.parent.focus_on(None)


class TimeAttrField(TextAttrField):

	def value_to_text(self, secs):
		return "%d:%02d" % (secs // 60, secs % 60)
	
	def text_to_value(self, x):
		t = "".join([c for c in x if c in ":0123456789"])
		if ":" not in t:
			t = ":" + t
		m, s = t.split(":")[:2]
		def i(v):
			return int(v or "0")
		return i(m) * 60 + i(s)


#----------------------------------------------------------------------------

class TextDisplay(Widget):

	fg_color = (255, 255, 255)

	def __init__(self, width, client, attr, font = None,
			click_action = None):
		font = font or get_default_font()
		if isinstance(width, str):
			size = font.size(width)
		else:
			size = width, font.size("X")[1]
		Widget.__init__(self, Rect((0, 0), size))
		self.font = font
		self.client = client
		self.attr = attr
		if click_action:
			self.click_action = click_action

	def draw(self, surf):
		frame = surf.get_rect()
		text = self.get_text()
		buf = self.font.render(text, True, self.fg_color)
		r = buf.get_rect()
		r.midleft = frame.midleft
		surf.blit(buf, r)
	
	def get_text(self):
		return getattr(self.client, self.attr)
	
	def mouse_down(self, e):
		self.call_handler('click_action')

#----------------------------------------------------------------------------

class TimeDisplay(TextDisplay):

	def __init__(self, client, attr, **kwds):
		TextDisplay.__init__(self, "12:59", client, attr,  **kwds)
	
	def get_text(self):
		return time_to_text(TextDisplay.get_text(self))

#----------------------------------------------------------------------------

def text_to_time(x):
	if not x:
		return -1
	else:
		t = "".join([c for c in x if c in ":0123456789"])
		if ":" not in t:
			t = ":" + t
		m, s = t.split(":")[:2]
		def i(v):
			return int(v or "0")
		return i(m) * 60 + i(s)

def time_to_text(secs):
	if secs < 0:
		return ""
	else:
		return "%d:%02d" % (secs // 60, secs % 60)

#----------------------------------------------------------------------------

class Form(Dialog):

	border_width = 1

	def __init__(self, *args, **kwds):
		Dialog.__init__(self, *args, **kwds)
		self.labels = []
		self.fields = []
		self.fields_by_name = {}
	
	def text_field(self, name, caption, *args, **kwds):
		self.add_field(name, caption, TextField(*args, **kwds))
	
	def add_field(self, name, caption, field):
		self.labels.append(Label(caption))
		self.fields.append(field)
		self.fields_by_name[name] = field
	
	def end(self):
		labels = Column(self.labels, align = 'l')
		fields = Column(self.fields, align = 'l')
		labels_and_fields = Row([labels, fields])
		buttons = Row([
			Button("OK", action = self.ok),
			Button("Cancel", action = self.cancel)
		])
		all = Column([labels_and_fields, buttons], align = 'r')
		all.rect.topleft = (10, 10)
		self.add(all)
		self.shrink_wrap()
		if self.fields:
			self.fields[0].focus()

	def ok(self):
		self.dismiss(True)
	
	def cancel(self):
		self.dismiss(False)

	def __getitem__(self, name):
		return self.fields_by_name[name].get_value()

	def __setitem__(self, name, value):
		self.fields_by_name[name].set_value(value)
