#
#   5B - Editor
#

from pygame.locals import *
from pygame import draw
from albow.widget import Widget
from albow.controls import Label, Button, Row, Column, Frame
from albow.palette_view import PaletteView
from albow.resource import get_font, get_image
from controls import TextDisplay, TimeDisplay, text_to_time, time_to_text, \
	Form
from screen import Screen, menu_button
from level_view import LevelView
from part import Part
from wire import Wire, WirePoint, get_wire_color, num_wire_colors, wire_width
from vector_algebra import Point2, IdentTran2, ScaleTran2, Zero2
import parts, settings

tool_palette_cell_size = (30, 30)
part_palette_width = 100
wire_palette_width = 100
wire_palette_item_height = 10
grid_size = 5

SELECT_TOOL = 0
ADD_TOOL = 1
WIRE_TOOL = 2

#----------------------------------------------------------------------------

class EditorMenuScreen(Screen):

	def __init__(self, shell):
		Screen.__init__(self, shell)
		game = shell.game
		title = Label("Level Editor", font = get_font(24, "VeraBd.ttf"))
		def mb(title, action, enable = None):
			return menu_button(title, action = action, enable = enable)
		items = [
			mb("New Level",   shell.edit_new_level),
			mb("Edit Level",  shell.show_editor_level, game.level_present),
			mb("Load Level",  shell.load_level),
			mb("Save Level", shell.save_level,        game.level_present),
			mb("Exit Editor", shell.show_menu),
		]
		menu = Column(items, align = 'l')
		box = Column([title, menu], align = 'l', spacing = 24)
		self.center(box)

#----------------------------------------------------------------------------

class EditorLevelScreen(Screen):

	def __init__(self, shell):
		Screen.__init__(self, shell)
		sw, sh = settings.screen_size
		level_view = EditorLevelView(Rect((0, 0), (sw, sh)), self.game)
		tool_palette = EditorToolPalette(level_view)
		wire_palette = WireColorPalette(level_view)
		part_palette = EditorPartPalette(level_view)
		buttons = self.make_buttons(level_view)
		#start_time_field = TimeAttrField(35, level_view, 'start_time')
		#deadline_field = TimeAttrField(35, level_view, 'deadline')
		#script_field = TextAttrField(100, level_view, 'script', upper = True)
		start_time_field = TimeDisplay(level_view, 'start_time',
			click_action = level_view.edit_start_time)
		deadline_field = TimeDisplay(level_view, 'deadline',
			click_action = level_view.edit_deadline)
		script_field = TextDisplay(100, level_view, 'script',
			click_action = level_view.edit_script)
		def f(w):
			return Frame(w, border_width = 1, border_spacing = 2)
		def lf(label, w):
			fw = f(w)
			l = Label(label)
			return Row([l, fw])
		fields = Column([
			lf("Init Time", start_time_field),
			lf("Deadline", deadline_field),
			Label("Tyger Script", width = 105),
			f(script_field),
		], align = 'r')
		box = Column(
			[tool_palette, wire_palette, part_palette, buttons, fields],
			height = sh, expand = part_palette)
		box.rect.topright = (sw, 0)
		level_view.rect.width = box.rect.left
		self.add(level_view)
		self.add(box)
		level_view.focus()
	
	def make_buttons(self, level_view):
		return Column([
			Button("Flip Horiz", action = level_view.flip_h),
			Button("Flip Vert", action = level_view.flip_v),
			Button("Rotate", action = level_view.rotate),
			Button("Edit Params", action = level_view.edit_params),
		], align = 'l', spacing = 0)
	
	def key_down(self, e):
		key = e.key
		if key == K_ESCAPE:
			self.shell.show_editor_menu()
		else:
			Screen.key_down(self, e)

#----------------------------------------------------------------------------

class EditorToolPalette(PaletteView):

	images = [
		get_image("editor-select.png"),
		get_image("editor-add.png"),
		get_image("editor-wire.png"),
	]
	
	highlight_style = 'fill'
	
	def __init__(self, editor):
		PaletteView.__init__(self, tool_palette_cell_size, 1, self.num_items())
		self.editor = editor
	
	def num_items(self):
		return len(self.images)
	
	def draw_item(self, surf, item_no, frame):
		image = self.images[item_no]
		r = image.get_rect()
		r.center = frame.center
		surf.blit(image, r)
	
	def click_item(self, item_no, e):
		self.editor.set_tool(item_no)
	
	def item_is_selected(self, item_no):
		return item_no == self.editor.tool

#----------------------------------------------------------------------------

class EditorPartPalette(PaletteView):

	font = get_font(12, "Vera.ttf")
	fg_color = (255, 255, 255)
	highlight_style = 'fill'

	def __init__(self, editor):
		cw = part_palette_width
		ch = self.font.get_linesize()
		PaletteView.__init__(self, (cw, ch), self.num_items(), 1)
		self.editor = editor
	
	def num_items(self):
		return len(parts.catalog)
	
	def draw_item(self, surf, item_no, frame):
		name = parts.catalog[item_no].name
		buf = self.font.render(name, True, self.fg_color)
		surf.blit(buf, frame)
	
	def click_item(self, item_no, e):
		self.editor.part_no = item_no
		self.editor.set_tool(ADD_TOOL)
	
	def item_is_selected(self, item_no):
		return item_no == self.editor.part_no

#----------------------------------------------------------------------------

class WireColorPalette(PaletteView):

	def __init__(self, editor):
		nr = (self.num_items() + 1) // 2
		nc = 2
		cw = wire_palette_width // nc
		ch = wire_palette_item_height
		PaletteView.__init__(self, (cw, ch), nr, 2)
		self.editor = editor
	
	def num_items(self):
		return num_wire_colors()
		
	def draw_item(self, surf, item_no, frame):
		x, y = frame.topleft
		d = frame.height // 2 - 1
		draw.line(surf, get_wire_color(item_no),
			(x + d, y + d),
			(x + frame.width - d, y + d),
			wire_width)

	def click_item(self, item_no, e):
		self.editor.wire_color_no = item_no
		self.editor.set_tool(WIRE_TOOL)
	
	def item_is_selected(self, item_no):
		return item_no == self.editor.wire_color_no

#----------------------------------------------------------------------------

class EditorLevelView(LevelView):
	#  tool                 int
	#  part_no              int
	#  wire_color_no        int
	#  click_point          Point2
	#  rubber_point         Point2
	#  editing_wire_point   WirePoint or None
	#  dragged              boolean

	highlight_color = (0, 128, 255)

	def __init__(self, rect, game):
		LevelView.__init__(self, rect, game)
		self.tool = SELECT_TOOL
		self.part_no = 0
		self.wire_color_no = 0
		self.click_point = None
		self.rubber_point = None
		self.editing_wire_point = None
		self.dragged = False

	level = property(lambda self: self.game.current_level)
	
	def set_tool(self, tool):
		if tool <> self.tool:
			if self.tool == WIRE_TOOL:
				self.finish_wire()
		self.tool = tool
		if tool <> SELECT_TOOL:
			self.game.current_level.clear_selection()
	
	def draw_with_transform(self, surf, t, with_selection):
		LevelView.draw_with_transform(self, surf, t, with_selection = True)
		level = self.level
		for item in level.selection:
			if isinstance(item, WirePoint):
				item.draw_highlight(surf, t, self.highlight_color)
		if self.tool == WIRE_TOOL:
			wp = self.editing_wire_point
			if wp:
				p = self.rubber_point
				if p:
					wp.draw_rubber(surf, t, p)
	
	def mouse_down(self, e):
		self.dragged = False
		self.alt_clicked = e.alt
		tool = self.tool
		if tool == SELECT_TOOL:
			self.select_tool_click(e)
		elif tool == ADD_TOOL:
			self.add_tool_click(e)
		elif tool == WIRE_TOOL:
			self.wire_tool_click(e)
	
	def select_tool_click(self, e):
		self.click_point = Point2(e.local)
		level = self.level
		hit = self.find_clicked_item(e, wires_only = e.alt)
		#print "editor: select_tool_click: hit =", hit ###
		if not e.shift and hit not in level.selection:
			level.clear_selection()
		if hit:
			if e.shift:
				level.toggle_selection(hit)
			else:
				level.add_to_selection(hit)

	def add_tool_click(self, e):
		#print "Editor: add_tool_click" ###
		part = self.instantiate_part()
		if part:
			#print "...part =", part ###
			p = self.mouse_to_global_snapped(e)
			#print "...positioning at", p ###
			part.position = p
			part.edit_params()
			self.level.add_part(part)
	
	def wire_tool_click(self, e):
		p = self.mouse_to_global_snapped(e)
		wp = self.editing_wire_point
		if not wp:
			wp = self.level.find_wire_point(p)
			if wp and not wp.connected_part:
				self.editing_wire_point = wp
			else:
				wire = Wire(self.wire_color_no, p)
				self.editing_wire_point = wire.corners[0]
			self.rubber_point = None
		else:
			level = self.level
			wire = wp.wire
			new_wire = wire.num_points() == 1
			wp = level.extend_wire_to_point(wp, p)
			self.editing_wire_point = wp
			if wp:
				if new_wire:
					level.add_wire(wire)
				if wp.connected_part:
					self.finish_wire()
	
	def mouse_drag(self, e):
		if self.tool == SELECT_TOOL:
			p = Point2(e.local)
			d = (p - self.click_point).snap(grid_size)
			if d <> Zero2:
				self.level.offset_selection(d,
					fork_points = not self.dragged and self.alt_clicked)
				self.click_point += d
				self.dragged = True
				self.invalidate()
	
	def mouse_move(self, e):
		if self.tool == WIRE_TOOL:
			if self.editing_wire_point:
				self.rubber_point = self.mouse_to_global_snapped(e)
				self.invalidate()
	
	def key_down(self, e):
		tool = self.tool
		key = e.key
		if key in (K_BACKSPACE, K_DELETE):
			self.level.delete_selection()
		elif key == K_ESCAPE and tool == WIRE_TOOL and self.editing_wire_point:
			print "Editor: Escaping from wire" ###
			self.finish_wire()
		else:
			Widget.key_down(self, e)
	
	def finish_wire(self):
		self.editing_wire_point = None
	
	def instantiate_part(self):
		i = self.part_no
		cat = parts.catalog
		#print "Editor: instantiate_part:", i ###
		if 0 <= i < len(cat):
			return cat[i].instantiate()
	
	def mouse_to_global_snapped(self, e):
		p = self.mouse_to_global(e)
		return p.snap(grid_size)

	def get_start_time(self):
		return self.level.start_time_secs
	
	def set_start_time(self, x):
		self.level.start_time_secs = x

	start_time = property(get_start_time, set_start_time)

	def get_deadline(self):
		return self.level.deadline_secs
	
	def set_deadline(self, x):
		self.level.deadline_secs = x

	deadline = property(get_deadline, set_deadline)

	def get_script(self):
		return self.level.get_script()
	
	def set_script(self, x):
		self.level.set_script(x)

	script = property(get_script, set_script)

	def edit_params(self):
		for part in self.level.selection:
			part.edit_params()
			return

	def flip_h(self):
		self.rot_flip(2)
	
	def flip_v(self):
		self.rot_flip(4)
	
	def rotate(self):
		self.rot_flip(1)
	
	def rot_flip(self, bits):
		self.level.rotflip_selection(bits)

	def edit_time(self, caption, attr):
		form = Form()
		form.text_field('time', caption, 40)
		form['time'] = time_to_text(getattr(self.level, attr))
		form.end()
		if form.present():
			setattr(self.level, attr, text_to_time(form['time']))

	def edit_start_time(self):
		self.edit_time("Start Time", 'start_time_secs')
	
	def edit_deadline(self):
		self.edit_time("Start Time", 'deadline_secs')

	def edit_script(self):
		form = Form()
		form.text_field('script', "Tyger Script", 500, upper = True)
		form['script'] = self.script
		form.end()
		if form.present():
			self.script = form['script']
